#!/usr/bin/opsi-python
# -*- coding: utf-8 -*-

# This module is part of the desktop management solution opsi
# (open pc server integration) http://www.opsi.org

# Copyright (C) 2010-2019 uib GmbH

# http://www.uib.de/

# All rights reserved.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""
Link drivers to the directories expected in the netboot product.

@copyright:     uib GmbH <info@uib.de>
@author: Jan Schneider <j.schneider@uib.de>
@author: Erol Ueluekmen <e.ueluekmen@uib.de>
@author: Niko Wenselowski <n.wenselowski@uib.de>
@license: GNU General Public License version 2
"""

from __future__ import print_function

import argparse
import grp
import os
import pwd
import re
import shutil
import sys

from OPSI.Config import (
        DEFAULT_DEPOT_USER as DEFAULT_CLIENT_DATA_USER,
        FILE_ADMIN_GROUP as DEFAULT_CLIENT_DATA_GROUP)
from OPSI.System import execute, which
from OPSI.Util import findFiles
from OPSI.Util.File import InfFile, PciidsFile, UsbidsFile

from opsicommon.logging import LOG_DEBUG, LOG_WARNING, logger, logging_config, DEFAULT_COLORED_FORMAT

try:
    from OPSI.Util.Path import cd
except ImportError:
    from contextlib import contextmanager

    @contextmanager
    def cd(path):
        'Change the current directory to `path` as long as the context exists.'

        currentDir = os.getcwd()
        os.chdir(path)
        try:
            yield
        finally:
            os.chdir(currentDir)

__version__ = '4.2.0.2'

DST_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'drivers')
PREFERRED_DRIVERS_DIRNAME = 'preferred'
ADDITIONAL_DRIVERS_DIRNAME = 'additional'
EXCLUDE_DRIVERS_DIRNAME = 'excluded'
BUILDIN_DRIVERS_DIRNAME = 'buildin'


def setRights():
    def isRoot():
        return os.geteuid() == 0

    if isRoot():
        # Only root can chown foreign users
        uid = pwd.getpwnam(DEFAULT_CLIENT_DATA_USER)[2]
    else:
        uid = -1

    gid = grp.getgrnam(DEFAULT_CLIENT_DATA_GROUP)[2]

    for path in findFiles(prefix=DST_DIR, directory=DST_DIR, returnLinks=False):
        logger.info("Setting owner of '%s' to '%s:%s'", path, uid, gid)
        try:
            os.chown(path, uid, gid)
        except Exception as e:
            logger.warning("Failed to change owner of '%s' to '%s:%s': %s", path, uid, gid, e)

        try:
            if os.path.islink(path):
                continue
            elif os.path.isdir(path):
                logger.debug("Setting rights on directory '%s'", path)
                os.chmod(path, 0o2770)
            elif os.path.isfile(path):
                logger.debug("Setting rights on file '%s'", path)
                os.chmod(path, (os.stat(path)[0] | 0o660) & 0o770)
        except Exception as e:
            logger.warning("Failed to set access rights of '%s' : %s", path, e)


def unlink():
    for dirname in ('pciids', 'usbids', 'hdaudioids', 'classes', 'vendors'):
        linkdir = os.path.join(DST_DIR, dirname)
        if os.path.exists(linkdir):
            logger.notice("Deleting link directory '%s'", linkdir)
            shutil.rmtree(linkdir)


def link():
    with cd(DST_DIR):
        if not os.path.exists("pci.ids"):
            logger.notice("Downloading pci.ids")
            execute("%s https://raw.githubusercontent.com/pciutils/pciids/master/pci.ids" % which('wget'))
        if not os.path.exists("usb.ids"):
            logger.notice("Downloading usb.ids")
            execute("%s http://www.linux-usb.org/usb.ids" % which('wget'))

        pciidsFile = PciidsFile("pci.ids")
        usbidsFile = UsbidsFile("usb.ids")

        logger.notice("Searching for inf files")
        if not os.path.isdir('drivers'):
            raise Exception(f"No drivers found, please copy your drivers to {os.path.abspath('drivers')}")

        driverDirs = os.listdir('drivers')  # this is drivers/drivers
        logger.debug("Found the following entries: %s", driverDirs)
        driverDirs = [
                directory for directory in driverDirs
                if os.path.isdir(os.path.join('drivers', directory))
        ]
        logger.debug("After filtering for directories: %s", driverDirs)

        infFiles = []
        for directory in driverDirs:
            if directory != PREFERRED_DRIVERS_DIRNAME:
                continue
            driverDir = os.path.join('drivers', directory)
            logger.debug("Searching for files in %r", driverDir)
            infFiles.extend(findFiles(prefix=driverDir, directory=driverDir, includeFile=re.compile('\.inf$', re.IGNORECASE), returnDirs=False, followLinks=True))

        for directory in driverDirs:
            if directory != BUILDIN_DRIVERS_DIRNAME:
                continue
            driverDir = os.path.join('drivers', directory)
            logger.debug("Searching for files in %r", driverDir)
            infFiles.extend(findFiles(prefix=driverDir, directory=driverDir, includeFile=re.compile('\.inf$', re.IGNORECASE), returnDirs=False, followLinks=True))

        for directory in driverDirs:
            if directory in (PREFERRED_DRIVERS_DIRNAME, BUILDIN_DRIVERS_DIRNAME, ADDITIONAL_DRIVERS_DIRNAME, EXCLUDE_DRIVERS_DIRNAME):
                continue
            driverDir = os.path.join('drivers', directory)
            logger.debug("Searching for files in %r", driverDir)
            infFiles.extend(findFiles(prefix=driverDir, directory=driverDir, includeFile=re.compile('\.inf$', re.IGNORECASE), returnDirs=False, followLinks=True))

        # TODO: was ist mit Dateien, welche direkt in drivers/drivers liegen?
        logger.debug("Found %s files: %s", len(infFiles), infFiles)

        logger.notice("Parsing inf files")
        for index, filename in enumerate(infFiles, start=1):
            print(f"\r\033[0K [{index}/{len(infFiles)}] {os.path.basename(filename)}\r")
            sys.stdout.flush()
            logger.debug("Processing %r", filename)
            infFile = InfFile(filename)
            for dev in infFile.getDevices():
                # Example dev:
                # {'device': '0268', 'path': None, 'vendor': '10EC',
                #  'class': 'media', 'type': 'HDAUDIO'}
                baseDir = ''
                idinfoFile = None
                if dev['type'] == 'PCI':
                    baseDir = 'pciids'
                    idinfoFile = pciidsFile
                elif dev['type'] == 'USB':
                    baseDir = 'usbids'
                    idinfoFile = usbidsFile
                elif dev['type'] == 'HDAUDIO':
                    baseDir = 'hdaudioids'
                elif dev['type'] == 'ACPI':
                    continue
                else:
                    logger.error(
                        "Unknown device type %s for device %s:%s, infFile: %s",
                        dev['type'], dev['vendor'], dev['device'], os.path.abspath(infFile.getFilename())
                    )
                    continue

                if not os.path.exists(baseDir):
                    os.mkdir(baseDir)
                if not os.path.exists(os.path.join(baseDir, dev['vendor'])):
                    os.mkdir(os.path.join(baseDir, dev['vendor']))

                if os.path.exists(os.path.join(baseDir, dev['vendor'], dev['device'])):
                    link = os.path.normpath(
                            os.path.join(
                                    'remove', 'remove',
                                    os.readlink(
                                            os.path.join(
                                                    baseDir, dev['vendor'], dev['device']
                                            )
                                    )
                            )
                    )

                    dirn = os.path.dirname(infFile.getFilename())
                    if link != dirn:
                        logger.info(
                            "Duplicate driver found for %s device %s:%s, infFile: %s",
                            dev['type'], dev['vendor'], dev['device'], os.path.abspath(infFile.getFilename())
                        )
                    continue

                path = os.path.join('..', '..', dev['path'])
                os.symlink(path, os.path.join(baseDir, dev['vendor'], dev['device']))

                if not idinfoFile:
                    continue

                vendorName = idinfoFile.getVendor(dev['vendor'])
                if not vendorName:
                    continue

                vendorName = vendorName.replace('/', '-')
                if not os.path.exists('vendors'):
                    os.mkdir('vendors')
                if not os.path.exists(os.path.join('vendors', vendorName)):
                    os.mkdir(os.path.join('vendors', vendorName))

                deviceName = idinfoFile.getDevice(dev['vendor'], dev['device'])
                if not deviceName:
                    continue

                deviceName = deviceName.replace('/', '-')
                if not os.path.exists(os.path.join('vendors', vendorName, deviceName)):
                    path = os.path.join('..', '..', dev['path'])
                    os.symlink(path, os.path.join('vendors', vendorName, deviceName))

                if not os.path.exists('classes'):
                    os.mkdir('classes')
                if not os.path.exists(os.path.join('classes', dev['class'])):
                    os.mkdir(os.path.join('classes', dev['class']))
                if not os.path.exists( os.path.join('classes', dev['class'], vendorName)):
                    os.mkdir(os.path.join('classes', dev['class'], vendorName))
                if not os.path.exists(os.path.join('classes', dev['class'], vendorName, deviceName)):
                    path = os.path.join('..', '..', '..', dev['path'])
                    os.symlink(path, os.path.join('classes', dev['class'], vendorName, deviceName))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Create links needed for automatic driver integration used in opsi-linux-bootimage")
    parser.add_argument('--version', '-V', action='version', version=__version__)
    parser.add_argument('--verbose', '-v',
                                            dest="logLevel", default=LOG_WARNING, action="count",
                                            help="increase verbosity (can be used multiple times)")
    parser.add_argument('--debug-file', dest='debugFile',
                                            help='Write debug output to given file.')

    args = parser.parse_args()

    logging_config(stderr_level=args.logLevel, file_level=LOG_DEBUG, log_file=args.debugFile, stderr_format=DEFAULT_COLORED_FORMAT)

    try:
        # check if program is available
        which('wget')

        unlink()
        link()
        setRights()
    except Exception as error:
        logger.error(error, exc_info=True)
        raise error
