#!/usr/bin/opsi-python
# -*- coding: utf-8 -*-
#
# This module is part of the desktop management solution opsi
# (open pc server integration) http://www.opsi.org
#
# Copyright (C) 2006-2010, 2013-2019 uib GmbH <info@uib.de>
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Show drivers that would be used for setting up a client.

:copyright:     uib GmbH <info@uib.de>
:author: Jan Schneider <j.schneider@uib.de>
:author: Erol Ueluekmen <e.ueluekmen@uib.de>
:author: Niko Wenselowski <n.wenselowski@uib.de>
:license: GNU Affero General Public License version 3
"""

from __future__ import print_function

import argparse
import os
import re
import socket

from OPSI.Logger import COLOR_NORMAL, COLOR_LIGHT_RED, COLOR_LIGHT_GREEN, COLOR_LIGHT_WHITE, COLOR_LIGHT_YELLOW
from OPSI.Types import forceHostId
from OPSI.Util import findFiles
from OPSI.Util.File import InfFile

from opsicommon.logging import logger, logging_config, LOG_WARNING, DEFAULT_COLORED_FORMAT
try:
    from opsicommon.client.opsiservice import get_service_client
    BackendManager = None
except (ModuleNotFoundError, ImportError):
    get_service_client = None
    from OPSI.Backend.BackendManager import BackendManager

__version__ = '4.2.0.3'

localDir = os.path.dirname(os.path.abspath(__file__))
pciids = os.path.join(localDir, 'drivers', 'pciids')
usbids = os.path.join(localDir, 'drivers', 'usbids')
hdaudioids = os.path.join(localDir, 'drivers', 'hdaudioids')
manually = os.path.join(localDir, 'drivers', 'drivers', 'additional')
byAuditmanually = os.path.join(localDir, 'drivers', 'drivers', 'additional', 'byAudit')
buildin = os.path.join(localDir, 'drivers', 'drivers', 'buildin')
productId = os.path.basename(os.path.dirname(os.path.abspath(__file__)))


def main(hostId):
    if hostId.count('.') < 2:
        hostId += '.' + '.'.join(socket.getfqdn().split('.')[1:])
    hostId = forceHostId(hostId)
    logger.notice("Using host id: %s", hostId)

    try:
        if get_service_client:
            backend = get_service_client()
        else:
            backend = BackendManager()
        try:
            host = backend.host_getObjects(id=hostId)[0]  # pylint: disable=no-member
        except IndexError:
            raise Exception(f"Host '{hostId}' not found in opsi database")

        auditHardwareOnHosts = backend.auditHardwareOnHost_getObjects(hostId=host.id, state=1)
        if not auditHardwareOnHosts:
            raise Exception(f"Hardware information for host '{host.id}' not found, please run 'hwinvent' or 'hwaudit' first.")

        productProperties = backend.productPropertyState_getObjects(objectId=host.id, productId=productId, propertyId="additional_drivers")
    finally:
        if get_service_client:
            backend.disconnect()
        else:
            backend.backend_exit()
    try:
        additionaldrivers = [
                ad.strip() for ad
                in ','.join(productProperties[0].getValues()).split(',')
                if ad.strip()
        ]
    except IndexError:  # no productProperties
        additionaldrivers = []

    byAuditIntegrated = False
    if os.path.exists(byAuditmanually):
        vendorFromHost = None
        modelFromHost = None
        skuFromHost = ""
        skuLabel = ""
        fallbackPath = ""

        for auditHardwareOnHost in auditHardwareOnHosts:
            if not auditHardwareOnHost.hardwareClass == "COMPUTER_SYSTEM":
                continue
            if auditHardwareOnHost.state != 1:
                continue

            vendorFromHost = re.sub(r'[\<\>\?"\:\|\\/\*]', "_", auditHardwareOnHost.vendor)
            modelFromHost = re.sub(r'[\<\>\?"\:\|\\/\*]', "_", auditHardwareOnHost.model)
            skuFromHost = auditHardwareOnHost.sku or ""

        if vendorFromHost and modelFromHost:
            print(
                yellow(
                    "Additional drivers for integration found using byAudit (System) for "
                    f"vendor: '{vendorFromHost}' model : '{modelFromHost}'  "
                    "Check if drivers are available."
                )
            )
            vendorDirectories = os.listdir(byAuditmanually)
            if vendorFromHost not in vendorDirectories:
                if vendorFromHost.endswith(".") or vendorFromHost.endswith(" "):
                    vendorFromHost = "%s_" % vendorFromHost[:-1]

            for vendorDirectory in vendorDirectories:
                if vendorDirectory.lower() == vendorFromHost.lower():
                    modelDirectories = os.listdir(os.path.join(byAuditmanually, vendorDirectory))
                    if skuFromHost and skuFromHost in modelFromHost:
                        skuLabel = "(%s)" % skuFromHost

                    for modelDirectory in modelDirectories:
                        if modelDirectory.lower() == modelFromHost.lower():
                            print(yellow("ByAudit: Exact match found."))
                            additionaldrivers.append(os.path.join(byAuditmanually, vendorDirectory, modelDirectory))
                            byAuditIntegrated = True
                            break
                        elif modelDirectory.lower() == modelFromHost.replace(skuLabel, "").strip().lower():
                            fallbackPath = os.path.join("byAudit", vendorDirectory, modelDirectory)

                    if not byAuditIntegrated and fallbackPath:
                        print(yellow(f"ByAudit: No Exact match found but model without sku found. Using Directory: '{fallbackPath}'"))
                        additionaldrivers.append(fallbackPath)
                        byAuditIntegrated = True

                    break

    if os.path.exists(byAuditmanually) and not byAuditIntegrated:
        vendorFromHost = None
        productFromHost = None
        for auditHardwareOnHost in auditHardwareOnHosts:
            if not auditHardwareOnHost.hardwareClass == "BASE_BOARD":
                continue
            if not auditHardwareOnHost.state == 1:
                continue
            vendorFromHost = re.sub(r'[\<\>\?"\:\|\\/\*]', "_", auditHardwareOnHost.vendor)
            productFromHost = re.sub(r'[\<\>\?"\:\|\\/\*]', "_", auditHardwareOnHost.product)

        if vendorFromHost and productFromHost:
            print(
                yellow(
                    "Additional drivers for integration found using byAudit (Motherboard) for "
                    f"vendor: '{vendorFromHost}' model : '{productFromHost}'  "
                    "Check if drivers are available."
                )
            )
            vendorDirectories = os.listdir(byAuditmanually)
            if vendorFromHost not in vendorDirectories:
                if vendorFromHost.endswith(".") or vendorFromHost.endswith(" "):
                    vendorFromHost = "%s_" % vendorFromHost[:-1]

            for vendorDirectory in vendorDirectories:
                if vendorDirectory.lower() == vendorFromHost.lower():
                    productDirectories = os.listdir(os.path.join(byAuditmanually, vendorDirectory))
                    for productDirectory in productDirectories:
                        if productDirectory.lower() == productFromHost.lower():
                            additionaldrivers.append(os.path.join(byAuditmanually, vendorDirectory, productDirectory))

    alreadyIntegrated = {}
    manuallySelectedDrivers = {}
    multipleManualSelectedDrivers = {}
    if additionaldrivers:
        print(white("Manually selected drivers (additional)"))
        for manuallySelectedDriver in additionaldrivers:
            manuallySelectedDriver = manuallySelectedDriver.strip()
            print(white(f"   [{manuallySelectedDriver}]"))
            if os.path.islink(os.path.join(manually, manuallySelectedDriver)):
                target = os.path.join(
                        os.path.dirname(manuallySelectedDriver),
                        os.readlink(manuallySelectedDriver)
                )
            else:
                target = os.path.join(manually, manuallySelectedDriver)
            target = os.path.abspath(target)

            infFiles = findFiles(
                    directory=target,
                    prefix=target,
                    includeFile=re.compile(r'\.inf$', re.IGNORECASE),
                    returnDirs=False,
                    followLinks=True
            )
            logger.debug("Found inf files: %s in dir '%s'", infFiles, target)
            if not infFiles:
                print(red("      No .inf files found"))
                continue

            for inf in infFiles:
                dirname = os.path.dirname(inf)
                textMode = any(
                        filename.lower() == 'txtsetup.oem'
                        for filename in os.listdir(dirname)
                )
                if not textMode:
                    textMode = any(
                            filename.lower() == 'txtsetup.oem'
                            for filename in os.listdir(os.path.dirname(dirname))
                    )

                infFile = InfFile(inf)
                sourceDisksNames = infFile.getSourceDisksNames()
                if textMode:
                    print(green(f"      [{inf}] (textmode capable)"))
                else:
                    print(green(f"      [{inf}]"))

                for dev in infFile.getDevices():
                    # mark this device as already integrated
                    if dev['type'] not in alreadyIntegrated:
                        alreadyIntegrated[dev['type']] = {}
                    if dev['vendor'] not in alreadyIntegrated[dev['type']]:
                        alreadyIntegrated[dev['type']][dev['vendor']] = {}
                    if dev['device'] not in alreadyIntegrated[dev['type']][dev['vendor']]:
                        alreadyIntegrated[dev['type']][dev['vendor']][dev['device']] = infFile.getFilename()

                    if dev['type'] not in manuallySelectedDrivers:
                        manuallySelectedDrivers[dev['type']] = {}
                    if dev['vendor'] not in manuallySelectedDrivers[dev['type']]:
                        manuallySelectedDrivers[dev['type']][dev['vendor']] = {}

                    if dev['device'] in manuallySelectedDrivers[dev['type']][dev['vendor']]:
                        # if manuallySelectedDrivers are the same path
                        # like actual, don't check for multiple select!
                        pathInManuallySelected = os.path.dirname(manuallySelectedDrivers[dev['type']][dev['vendor']][dev['device']][1])
                        if pathInManuallySelected == os.path.dirname(os.path.abspath(infFile.getFilename())):
                            continue

                        # build multipeManualSelectedList for device, if needed
                        if dev['type'] not in multipleManualSelectedDrivers:
                            multipleManualSelectedDrivers[dev['type']] = {}
                        if dev['vendor'] not in multipleManualSelectedDrivers[dev['type']]:
                            multipleManualSelectedDrivers[dev['type']][dev['vendor']] = {}
                        if dev['device'] not in multipleManualSelectedDrivers[dev['type']][dev['vendor']]:
                            multipleManualSelectedDrivers[dev['type']][dev['vendor']][dev['device']] = []

                        # try to detect multiple selection
                        infFileDir = os.path.dirname(os.path.abspath(infFile.getFilename()))
                        foundInMultipleList = any(
                                infFileDir == os.path.dirname(os.path.abspath(multipleManualSelectedDriver[1]))
                                for multipleManualSelectedDriver
                                in multipleManualSelectedDrivers[dev['type']][dev['vendor']][dev['device']]
                        )
                        if not foundInMultipleList:
                            multipleManualSelectedDrivers[dev['type']][dev['vendor']][dev['device']].append([manuallySelectedDriver, infFile.getFilename(), textMode, sourceDisksNames])

                        continue
                    manuallySelectedDrivers[dev['type']][dev['vendor']][dev['device']] = [manuallySelectedDriver, infFile.getFilename(), textMode, sourceDisksNames]
                    logger.debug(
                        "Would integrate driver for %s device %s:%s, infFile: %s",
                        dev['type'], dev['vendor'], dev['device'], os.path.abspath(infFile.getFilename())
                    )
                    try:
                        if logger.level > LOG_WARNING:
                            print(
                                white(
                                    f"         {dev['type']} [{dev['vendor']}:{dev['device']}] {', '.join(sourceDisksNames)}"
                                )
                            )
                    except:
                        # opsi log level order is reversed
                        if logger.getConsoleLevel() < LOG_WARNING:
                            print(
                                white(
                                    f"         {dev['type']} [{dev['vendor']}:{dev['device']}] {', '.join(sourceDisksNames)}"
                                )
                            )
    print("")

    for hardwareClass in ('PCI_DEVICE', 'USB_DEVICE', 'HDAUDIO_DEVICE'):
        baseDir = None
        deviceType = None
        if hardwareClass == 'PCI_DEVICE':
            deviceType = 'PCI'
            baseDir = pciids
            print(white("PCI-Devices"))
        elif hardwareClass == 'USB_DEVICE':
            deviceType = 'USB'
            baseDir = usbids
            print(white("USB-Devices"))
        elif hardwareClass == 'HDAUDIO_DEVICE':
            deviceType = 'HDAUDIO'
            baseDir = hdaudioids
            print(white("HD-Audio-Devices"))

        devicesFound = 0
        for ah in auditHardwareOnHosts:
            if not (ah.hardwareClass == hardwareClass):
                continue
            if not ah.vendorId or not ah.deviceId:
                continue

            if ah.vendorId == '1D6B':  # Linux Foundation
                continue

            devicesFound += 1
            vendor = 'Unknown vendor'
            if hasattr(ah, 'vendor') and ah.vendor:
                vendor = ah.vendor
            name = 'Unknown device'
            if hasattr(ah, 'name') and ah.name:
                name = ah.name

            if hardwareClass == 'HDAUDIO_DEVICE':
                print(white(f"   [{ah.vendorId}:{ah.deviceId}]  {name}"))
            else:
                print(white(f"   [{ah.vendorId}:{ah.deviceId}]  {vendor} : {name}"))

            (manuallySelectedDriver, infFile, textMode, sourceDisksNames) = manuallySelectedDrivers.get(deviceType, {}).get(ah.vendorId, {}).get(ah.deviceId, [None, None, False, []])
            multipleSelectedDriverList = multipleManualSelectedDrivers.get(deviceType, {}).get(ah.vendorId, {}).get(ah.deviceId, [])
            if manuallySelectedDriver:
                if textMode:
                    print(
                        yellow(
                            f"      Manually selected [{manuallySelectedDriver}] {os.path.dirname(infFile)} (textmode capable)"
                        )
                    )
                else:
                    print(
                        yellow(
                            f"      Manually selected [{manuallySelectedDriver}] {os.path.dirname(infFile)}"
                        )
                    )
                if multipleSelectedDriverList:
                    for multipleSelectedDriver in multipleSelectedDriverList:
                        if multipleSelectedDriver[2]:
                            print(
                                yellow(
                                    f"      Multiple selected [{multipleSelectedDriver[0]}] {os.path.dirname(multipleSelectedDriver[1])} (textmode capable)"
                                )
                            )
                        else:
                            print(
                                yellow(
                                    f"      Multiple selected [{multipleSelectedDriver[0]}] {os.path.dirname(multipleSelectedDriver[1])}"
                                )
                            )
                continue

            vendorDriverDir = os.path.join(baseDir, ah.vendorId)
            if not os.path.exists(vendorDriverDir):
                print(red(f"      No driver - vendor directory '{vendorDriverDir}' not found"))
                continue

            driverDeviceDir = os.path.join(baseDir, ah.vendorId, ah.deviceId)
            if not os.path.exists(driverDeviceDir):
                print(red(f"      driver - device directory '{driverDeviceDir}' not found"))
                continue

            target = os.path.abspath( os.path.join(baseDir, ah.vendorId, os.readlink( os.path.join(baseDir, ah.vendorId, ah.deviceId) ) ) )
            if not target == buildin:
                infFiles = findFiles(
                        directory=target,
                        prefix=target,
                        includeFile=re.compile(r'\.inf$', re.IGNORECASE),
                        returnDirs=False,
                        followLinks=True
                )
                logger.debug("Found inf files: %s in dir '%s'", infFiles, target)
                for inf in infFiles:
                    infFile = InfFile(inf)
                    for dev in infFile.getDevices():
                        if dev['type'] not in alreadyIntegrated:
                            alreadyIntegrated[dev['type']] = {}
                        if dev['vendor'] not in alreadyIntegrated[dev['type']]:
                            alreadyIntegrated[dev['type']][dev['vendor']] = {}
                        if dev['device'] not in alreadyIntegrated[dev['type']][dev['vendor']]:
                            alreadyIntegrated[dev['type']][dev['vendor']][dev['device']] = infFile.getFilename()

            textMode = textMode or any(
                    filename.lower() == 'txtsetup.oem'
                    for filename in os.listdir(target)
            )
            if not textMode:
                textMode = any(
                        filename.lower() == 'txtsetup.oem'
                        for filename in os.listdir(os.path.dirname(target))
                )

            if target == buildin:
                print(yellow("      Using built-in windows driver"))
            elif textMode:
                print(green(f"      {target} (textmode capable)"))
            else:
                print(green(f"      {target}"))

            alreadyFoundInInfFile = alreadyIntegrated.get(deviceType, {}).get(ah.vendorId, {}).get(ah.deviceId, "")
            if alreadyFoundInInfFile and not (os.path.dirname(alreadyFoundInInfFile) == target):
                print(red(f"      This driver will not be integrated, because same device already integrated in: '{alreadyFoundInInfFile}'"))

        if not devicesFound:
            print(white("   No devices installed"))

        print("")


def yellow(text):
    return "%s%s%s" % (COLOR_LIGHT_YELLOW, text, COLOR_NORMAL)


def green(text):
    return "%s%s%s" % (COLOR_LIGHT_GREEN, text, COLOR_NORMAL)


def red(text):
    return '%s%s%s' % (COLOR_LIGHT_RED, text, COLOR_NORMAL)


def white(text):
    return '%s%s%s' % (COLOR_LIGHT_WHITE, text, COLOR_NORMAL)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Show which drivers will be integrated for the given host")
    parser.add_argument('--version', '-V', action='version', version=__version__)
    parser.add_argument('--verbose', '-v',
                                            dest="logLevel", default=LOG_WARNING, action="count",
                                            help="increase verbosity (can be used multiple times)")
    parser.add_argument('hostId')

    args = parser.parse_args()

    logging_config(stderr_level=args.logLevel, stderr_format=DEFAULT_COLORED_FORMAT)

    try:
        main(args.hostId)
    except Exception as error:
        logger.error(error, exc_info=True)
        raise error
